"""
AddCubeOperation
AddPlaneOperation
AddCylinderOperation
AddIcosahedronOperation
AddConeOperation
AddSphereOperation
AddTorusOperation
AddArrowOperation
"""


class MeshOperation:
    def __init__(self, base_widget):
        self.base_widget = base_widget

    def add_cube(self, **kwargs):
        self.base_widget.send_js_task(
            {"name": "ops.mesh.AddCubeOperation", "kwargs": kwargs}
        )

    def add_plane(self, **kwargs):
        self.base_widget.send_js_task(
            {"name": "ops.mesh.AddPlaneOperation", "kwargs": kwargs}
        )

    def add_cylinder(self, **kwargs):
        self.base_widget.send_js_task(
            {"name": "ops.mesh.AddCylinderOperation", "kwargs": kwargs}
        )

    def add_icosahedron(self, **kwargs):
        self.base_widget.send_js_task(
            {"name": "ops.mesh.AddIcosahedronOperation", "kwargs": kwargs}
        )

    def add_cone(self, **kwargs):
        self.base_widget.send_js_task(
            {"name": "ops.mesh.AddConeOperation", "kwargs": kwargs}
        )

    def add_sphere(self, **kwargs):
        self.base_widget.send_js_task(
            {"name": "ops.mesh.AddSphereOperation", "kwargs": kwargs}
        )

    def add_torus(self, **kwargs):
        self.base_widget.send_js_task(
            {"name": "ops.mesh.AddTorusOperation", "kwargs": kwargs}
        )

    def add_arrow(self, **kwargs):
        self.base_widget.send_js_task(
            {"name": "ops.mesh.AddArrowOperation", "kwargs": kwargs}
        )



"""
DeleteOperation
CopyOperation
"""


class ObjectOperation:
    def __init__(self, base_widget):
        self.base_widget = base_widget

    def delete(self, **kwargs):
        self.base_widget.send_js_task(
            {"name": "ops.object.DeleteOperation", "kwargs": kwargs}
        )

    def copy(self, **kwargs):
        self.base_widget.send_js_task(
            {"name": "ops.object.CopyOperation", "kwargs": kwargs}
        )



"""
# weas/src/operation/selection.js
This module contains the Transform class which is used to perform
SelectAll
InvertSelection
InsideSelection
"""


class SelectionOperation:
    def __init__(self, base_widget):
        self.base_widget = base_widget

    def select_all(self):
        self.base_widget.send_js_task({"name": "ops.selection.SelectAll", "kwargs": {}})

    def invert_selection(self, **kwargs):
        self.base_widget.send_js_task(
            {"name": "ops.selection.InvertSelection", "kwargs": kwargs}
        )

    def inside_selection(self, **kwargs):
        self.base_widget.send_js_task(
            {"name": "ops.selection.InsideSelection", "kwargs": kwargs}
        )



import numpy as np
from ase import Atoms
import cmath
from typing import List, Optional, Tuple


class Phonon:
    """
    A class to represent phonon vibrations and trajectories for a given atomic structure.

    Attributes:
        atoms (Atoms): The atomic structure.
        kpoint (Optional[List[float]]): The k-point in reciprocal space.
        eigenvectors (Optional[List[List[Tuple[float, float]]]]): Eigenvectors representing phonon modes.
        add_atom_phase (bool): Whether to add phase based on atom positions.
        vibrations (List[List[complex]]): Calculated vibration phases for each atom.
    """

    def __init__(
        self,
        atoms: Atoms,
        kpoint: Optional[List[float]] = None,
        eigenvectors: Optional[List[List[Tuple[float, float]]]] = None,
        add_atom_phase: bool = True,
    ):
        """
        Initialize the Phonon object.

        Args:
            atoms (Atoms): The atomic structure.
            kpoint (Optional[List[float]]): The k-point in reciprocal space.
            eigenvectors (Optional[List[List[Tuple[float, float]]]]): Eigenvectors for phonon modes.
            add_atom_phase (bool): Whether to add phase based on atom positions.
        """
        self.atoms = atoms
        self.kpoint = kpoint or [0, 0, 0]
        self.eigenvectors = eigenvectors
        self.add_atom_phase = add_atom_phase
        self.vibrations = []

    def calculate_vibrations(self, repeat: Tuple[int, int, int] = (1, 1, 1)) -> None:
        """
        Compute the initial phases and vibrations for the phonon modes.

        Args:
            repeat (Tuple[int, int, int]): Repeat factors for the atomic cell in x, y, and z directions.
        """
        nx, ny, nz = repeat
        fractional_positions = self.atoms.get_scaled_positions()
        natoms = len(self.atoms.positions)
        atom_phase = []

        if self.add_atom_phase:
            atom_phase = [np.dot(self.kpoint, pos) for pos in fractional_positions]
        else:
            atom_phase = [0] * natoms

        for ix in range(nx):
            for iy in range(ny):
                for iz in range(nz):
                    for i in range(natoms):
                        sprod = np.dot(self.kpoint, [ix, iy, iz]) + atom_phase[i]
                        phase = cmath.rect(1.0, sprod * 2.0 * np.pi)
                        self.vibrations.append(
                            [
                                phase * complex(vector[0], vector[1])
                                for vector in self.eigenvectors[i]
                            ]
                        )

    def get_trajectory(
        self,
        amplitude: float,
        nframes: int,
        kpoint: Optional[List[float]] = None,
        eigenvectors: Optional[List[List[Tuple[float, float]]]] = None,
        atoms: Optional[Atoms] = None,
        repeat: Tuple[int, int, int] = None,
        add_atom_phase: Optional[bool] = None,
    ) -> List[Atoms]:
        """
        Generate the trajectory of the phonon mode over time.

        Args:
            amplitude (float): Amplitude of the vibration.
            nframes (int): Number of frames for the trajectory.
            kpoint (Optional[List[float]]): K-point in reciprocal space.
            eigenvectors (Optional[List[List[Tuple[float, float]]]]): Eigenvectors for phonon modes.
            atoms (Optional[Atoms]): Atomic structure.
            repeat (Tuple[int, int, int]): Repeat factors for the atomic cell in x, y, and z directions.
            add_atom_phase (Optional[bool]): Whether to add phase based on atom positions.

        Returns:
            List[Atoms]: List of Atoms objects representing the trajectory frames.
        """
        if atoms is not None:
            self.atoms = atoms
        if kpoint is not None:
            self.kpoint = kpoint
        if eigenvectors is not None:
            self.eigenvectors = eigenvectors
        if self.kpoint is None or self.eigenvectors is None:
            raise ValueError("kpoint and eigenvectors must be provided")
        if add_atom_phase is not None:
            self.add_atom_phase = add_atom_phase

        repeat = repeat or [1, 1, 1]
        self.calculate_vibrations(repeat)
        trajectory = []
        times = [2 * np.pi * (i / nframes) for i in range(nframes)]

        for t in times:
            new_atoms = self.atoms * repeat
            phase = cmath.rect(amplitude, t)
            movement = []

            for i in range(len(new_atoms.positions)):
                displacement = [phase * v for v in self.vibrations[i]]
                displacement_real = [d.real for d in displacement]
                new_atoms.positions[i] = [
                    pos + disp / 5
                    for pos, disp in zip(new_atoms.positions[i], displacement_real)
                ]
                movement.append(displacement_real)

            new_atoms.set_array("movement", np.array(movement))
            trajectory.append(new_atoms)

        return trajectory


def generate_phonon_trajectory(
    atoms: Atoms,
    eigenvectors: List[List[Tuple[float, float]]],
    kpoint: List[float] = None,
    amplitude: float = 1,
    nframes: int = 20,
    repeat: Tuple[int, int, int] = None,
):
    """Generate a trajectory of atoms vibrating along the given eigenvectors.
    Args:
        atoms: ASE Atoms object
        eigenvectors: Eigenvectors of the dynamical matrix
        kpoint (Optional[List[float]]): K-point in reciprocal space.
        amplitude: Amplitude of the vibration
        nframes (int): Number of frames for the trajectory.
    Returns:
        A list of ASE Atoms objects representing the trajectory
    """
    phonon = Phonon(atoms, eigenvectors=eigenvectors)
    trajectory = phonon.get_trajectory(amplitude, nframes, kpoint=kpoint, repeat=repeat)
    return trajectory



from ..base_class import WidgetWrapper


class InstancedMeshPrimitive(WidgetWrapper):

    catalog = "instanced_mesh_primitive"

    _attribute_map = {
        "settings": "instancedMeshPrimitive",
    }

    _extra_allowed_attrs = []

    def __init__(self, _widget):
        super().__init__(_widget)



import difflib
from functools import wraps


class WidgetWrapper:
    """A base class for widget-related functionalities.

    This class provides dynamic property creation, attribute suggestions, and custom task application.
    """

    catalog = "base"
    _attribute_map = {}
    _extra_allowed_attrs = []

    def __init__(self, _widget):
        object.__setattr__(self, "_widget", _widget)
        for attr, widget_attr in self._attribute_map.items():
            self._create_property(attr, widget_attr)
        self._widget.observe(self._apply_task, "python_task")

    def _apply_task(self, change):
        if not change["new"] or change["new"]["catalog"] != self.catalog:
            return
        action = change["new"]["action"]
        setattr(self, action, change["new"]["data"])

    def _create_property(self, attribute_name, widget_attribute):
        def getter(self):
            return getattr(self._widget, widget_attribute)

        def setter(self, value):
            setattr(self._widget, widget_attribute, value)

        setattr(self.__class__, attribute_name, property(getter, setter))

    @property
    def all_attributes(self):
        return (
            list(self._attribute_map.keys())
            + ["_widget", "_attributes"]
            + self._extra_allowed_attrs
        )

    def _suggest_attribute(self, name):

        suggestions = difflib.get_close_matches(name, self.all_attributes)
        if suggestions:
            return f" Did you mean: {', '.join(suggestions)}?"
        else:
            return f" Available attributes are: {', '.join(self.all_attributes)}."

    def __setattr__(self, name, value):
        if name in self.all_attributes:
            object.__setattr__(self, name, value)
        else:
            suggestion = self._suggest_attribute(name)
            raise AttributeError(
                f"'{name}' is not a valid attribute of '{self.__class__.__name__}'.{suggestion}"
            )


def with_on_change(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        result = method(self, *args, **kwargs)  # Call the original method
        self._on_change()  # Trigger change tracking
        return result

    return wrapper


class ChangeTrackingDict(dict):
    def __init__(self, *args, widget=None, key=None, parent=None, **kwargs):
        self._changed = False
        self._parent = parent
        self._widget = widget
        self._key = key
        super().__init__(*args, **kwargs)

        # Wrap any nested dictionaries
        for k, value in self.items():
            if isinstance(value, dict):
                nested_dict = ChangeTrackingDict(value, parent=self)
                super().__setitem__(k, nested_dict)
        self._mark_changed()

    def __setitem__(self, key, value):
        if isinstance(value, dict) and not isinstance(value, ChangeTrackingDict):
            value = ChangeTrackingDict(value, parent=self)
            value._parent = self
        super().__setitem__(key, value)
        self._mark_changed()

    def __delitem__(self, key):
        super().__delitem__(key)
        self._mark_changed()

    def clear(self):
        super().clear()
        self._mark_changed()

    def pop(self, key, default=None):
        data = super().pop(key, default)
        self._mark_changed()
        return data

    def popitem(self):
        data = super().popitem()
        self._mark_changed()
        return data

    def update(self, *args, **kwargs):
        super().update(*args, **kwargs)
        self._mark_changed()

    def _mark_changed(self):
        """Set the changed flag to True, notify parent, and update widget if set."""
        self._changed = True
        if self._parent:
            self._parent._mark_changed()
        else:
            if not (self._widget and self._key):
                raise ValueError("Widget and key must be set to update the widget.")
            setattr(self._widget, self._key, self.as_dict())

    def has_changed(self):
        return self._changed

    def reset_changed_flag(self):
        """Reset the changed flag in this and all nested ChangeTrackingDicts."""
        self._changed = False
        for value in self.values():
            if isinstance(value, ChangeTrackingDict):
                value.reset_changed_flag()

    def as_dict(self):
        """Recursively convert ChangeTrackingDict to a standard dict."""
        result = {}
        for key, value in self.items():
            if isinstance(value, ChangeTrackingDict):
                result[key] = value.as_dict()
            else:
                result[key] = value
        return result



from ..base_class import WidgetWrapper, ChangeTrackingDict


class VectorField(WidgetWrapper):

    catalog = "vector_field"

    _attribute_map = {
        "show": "showVectorField",
    }

    _extra_allowed_attrs = ["_settings", "settings"]

    def __init__(self, _widget):
        super().__init__(_widget)

    @property
    def settings(self):
        return self._settings

    @settings.setter
    def settings(self, value):
        self._settings = ChangeTrackingDict(
            value, widget=self._widget, key="vectorField"
        )

    def update_atoms(self):
        self.settings = self.set_moment()

    def set_moment(self):
        """Set magnetic moment vector field."""
        atoms = self._widget.atoms
        if isinstance(atoms, list):
            atoms = atoms[0]
        settings = {}
        # separate spin up and down, add two vector fields
        if "moment" in atoms["attributes"]["atom"]:
            moment = atoms["attributes"]["atom"]["moment"]
            spin_up = [i for i, m in enumerate(moment) if m > 0]
            spin_down = [i for i, m in enumerate(moment) if m < 0]
            settings = {
                "up": {
                    "origins": [atoms["positions"][i] for i in spin_up],
                    "vectors": [[0, 0, moment[i]] for i in spin_up],
                    "color": "blue",
                },
                "down": {
                    "origins": [atoms["positions"][i] for i in spin_down],
                    "vectors": [[0, 0, moment[i]] for i in spin_down],
                    "color": "red",
                },
            }
        return settings



from ..base_class import WidgetWrapper, ChangeTrackingDict
from ase.data import atomic_numbers
from weas_widget.data import color_data, radii_data


class SpeciesManager(WidgetWrapper):

    catalog = "species"

    _attribute_map = {}

    _extra_allowed_attrs = ["_settings", "settings"]

    def __init__(self, _widget):
        super().__init__(_widget)
        self._settings = ChangeTrackingDict(widget=self._widget, key="speciesSettings")

    @property
    def settings(self):
        return self._settings

    @settings.setter
    def settings(self, value):
        self._settings = ChangeTrackingDict(
            value, widget=self._widget, key="speciesSettings"
        )

    def update_atoms(self):
        self.settings = self.get_default_settings()

    def get_default_settings(self):
        settings = {}
        atoms = self._widget.atoms
        if isinstance(atoms, list):
            atoms = atoms[0]
        species_dict = atoms.get("species", {})
        for species in species_dict:
            element = species_dict[species]
            number = atomic_numbers[element]
            color = list(color_data[self._widget.colorType][number])
            settings[species] = {
                "element": element,
                "symbol": species,
                "color": color,
                "radius": radii_data[self._widget.radiusType.upper()][number],
            }
        return settings



from ..base_class import WidgetWrapper
import numpy as np


class LatticePlane(WidgetWrapper):

    catalog = "lattice_plane"

    _attribute_map = {
        "planes": "anyMesh",
    }

    _extra_allowed_attrs = ["settings"]

    def __init__(self, _widget):
        super().__init__(_widget)
        self.settings = {}

    @property
    def atoms(self):
        return self._widget.atoms

    @property
    def cell(self):
        return np.array(self.atoms["cell"]).reshape(3, 3)

    @property
    def cell_edges(self):
        """Edges of the cell"""
        edge_indices = [
            [0, 3],
            [0, 1],
            [4, 2],
            [4, 1],
            [3, 5],
            [2, 6],
            [7, 5],
            [7, 6],
            [0, 2],
            [3, 6],
            [1, 5],
            [4, 7],
        ]
        basis = np.array(
            [
                [0, 0, 0],
                [1, 0, 0],
                [0, 1, 0],
                [0, 0, 1],
                [1, 1, 0],
                [1, 0, 1],
                [0, 1, 1],
                [1, 1, 1],
            ]
        )
        positions = np.dot(basis, self.cell)
        edges = []
        for indices in edge_indices:
            edges.append(positions[indices])
        return edges

    @property
    def cell_volume(self):
        return np.dot(self.cell[0], np.cross(self.cell[1], self.cell[2]))

    @property
    def cell_reciprocal(self):
        from math import pi

        b1 = 2 * pi / self.cell_volume * np.cross(self.cell[1], self.cell[2])
        b2 = 2 * pi / self.cell_volume * np.cross(self.cell[2], self.cell[0])
        b3 = 2 * pi / self.cell_volume * np.cross(self.cell[0], self.cell[1])
        return np.array([b1, b2, b3])

    def add_plane_from_indices(self, name, indices, distance=1.0, **kwargs):
        """Add a plane setting from indices"""
        cell_reciprocal = self.cell_reciprocal
        normal = np.dot(indices, cell_reciprocal)
        normal = normal / np.linalg.norm(normal)
        point = distance * normal
        self.settings[name] = {
            "normal": normal,
            "point": point,
        }
        self.settings[name].update(kwargs)

    def add_plane_from_selected_atoms(self, name, **kwargs):
        """
        Add a plane setting from selected atoms
        """
        indices = self._widget.selectedAtomsIndices
        if len(indices) != 3:
            raise ValueError("Please select three atoms.")
        positions = np.array(self.atoms["positions"])
        center = np.mean(positions[indices], axis=0)
        normal = np.cross(
            positions[indices[1]] - positions[indices[0]],
            positions[indices[2]] - positions[indices[0]],
        )
        self.settings[name] = {
            "normal": normal,
            "point": center,
        }
        self.settings[name].update(kwargs)

    def draw(self, no=None):
        """Draw plane
        no: int
            spacegroup of structure, if None, no will be determined by
            get_spacegroup_number()
        include_center: bool
            include center of plane in the mesh
        """

        # TODO delete old plane
        if no is not None:
            self.no = no
        planes = self.build_plane(self.cell)
        for name, plane in planes.items():
            self.send_js_task(
                {
                    "name": "drawPlane",
                    "kwargs": {"plane": plane},
                }
            )

    def build_plane(self):
        """
        Build vertices, edges and faces of plane.
        """
        cellEdges = self.cell_edges
        planes = {}
        for name, data in self.settings.items():
            intersect_points = []
            # get intersection point
            for line in cellEdges:
                intersect_point = linePlaneIntersection(
                    line, data["normal"], data["point"]
                )
                if intersect_point is not None:
                    intersect_points.append(intersect_point)
                # get verts, edges, faces by Hull
            if len(intersect_points) < 3:
                continue
            vertices, edges, faces = faces_from_vertices(
                intersect_points, data["normal"], scale=data.get("scale", 1)
            )
            planes[name] = self.get_plane_data(vertices, edges, faces, data)
        self.planes = list(planes.values())

    def get_plane_data(self, vertices, edges, faces, plane):
        """
        build edge
        """
        if len(faces) > 0:
            plane.update(
                {
                    "vertices": vertices.reshape(-1).tolist(),
                    "edges": edges,
                    "faces": np.array(faces).reshape(-1).tolist(),
                    "edges_cylinder": {
                        "lengths": [],
                        "centers": [],
                        "normals": [],
                        "vertices": 6,
                        "color": (0.0, 0.0, 0.0, 1.0),
                        "width": plane.pop("width", 0.1),
                        "battr_inputs": {},
                    },
                }
            )
            for edge in edges:
                center = (vertices[edge[0]] + vertices[edge[1]]) / 2.0
                vec = vertices[edge[0]] - vertices[edge[1]]
                length = np.linalg.norm(vec)
                nvec = vec / length
                plane["edges_cylinder"]["lengths"].append(length)
                plane["edges_cylinder"]["centers"].append(center)
                plane["edges_cylinder"]["normals"].append(nvec)
        return plane


def faces_from_vertices(vertices, normal, scale=[1, 1, 1]):
    """
    get faces from vertices
    """
    # remove duplicative point
    vertices = np.unique(vertices, axis=0)
    n = len(vertices)
    if n < 3:
        return vertices, [], []
    center = np.mean(vertices, axis=0)
    v1 = vertices[0] - center
    angles = [[0, 0]]
    normal = normal / (np.linalg.norm(normal) + 1e-6)
    for i in range(1, n):
        v2 = vertices[i] - center
        x = np.cross(v1, v2)
        c = np.sign(np.dot(x, normal))
        angle = np.arctan2(c, np.dot(v1, v2))
        angles.append([i, angle])
    # scale
    vec = vertices - center
    # length = np.linalg.norm(vec, axis = 1)
    # nvec = vec/length[:, None]
    vertices = center + np.array([scale]) * vec
    # search convex polyhedra
    angles = sorted(angles, key=lambda x: x[1])
    faces = []
    # change faces to triangle
    for i in range(1, n - 1):
        faces.append([angles[0][0], angles[i][0], angles[i + 1][0]])
    # get edges
    edges = []
    for i in range(n - 1):
        edges.append([angles[i][0], angles[i + 1][0]])
    return vertices, edges, faces


def linePlaneIntersection(line, normal, point):
    """
    3D Line Segment and Plane Intersection
    - Point
    - Line contained in plane
    - No intersection
    """
    d = np.dot(point, normal)
    normalLine = line[0] - line[1]
    a = np.dot(normalLine, normal)
    # No intersection or Line contained in plane
    if np.isclose(a, 0):
        return None
    # in same side
    b = np.dot(line, normal) - d
    if b[0] * b[1] > 0:
        return None
    # Point
    v = point - line[0]
    d = np.dot(v, normal) / a
    point = np.round(line[0] + normalLine * d, 6)
    return point



from .base_widget import BaseWidget
from .utils import ASEAdapter, PymatgenAdapter, load_online_example
from .atoms_viewer import AtomsViewer
from .camera import Camera
from .plugins.instanced_mesh_pritimive import InstancedMeshPrimitive
from .operators.ops import Operators
import time
import threading
import ipywidgets as ipw


class WeasWidget(ipw.HBox):
    def __init__(self, from_ase=None, from_pymatgen=None, **kwargs):
        self._widget = BaseWidget(**kwargs)
        super().__init__([self._widget])
        self.avr = AtomsViewer(self._widget)
        self.camera = Camera(self._widget)
        self.imp = InstancedMeshPrimitive(self._widget)
        self.ops = Operators(self._widget)
        if from_ase is not None:
            self.from_ase(from_ase)
        if from_pymatgen is not None:
            self.from_pymatgen(from_pymatgen)

    def from_ase(self, atoms):
        self.avr.atoms = ASEAdapter.to_weas(atoms)

    def to_ase(self):
        return ASEAdapter.to_ase(self.avr.atoms)

    def from_pymatgen(self, structure):
        self.avr.atoms = PymatgenAdapter.to_weas(structure)

    def to_pymatgen(self):
        return PymatgenAdapter.to_pymatgen(self.avr.atoms)

    def load_example(self, name="tio2.cif"):
        atoms = load_online_example(name)
        self.avr.atoms = ASEAdapter.to_weas(atoms)

    def export_image(self, resolutionScale=5):
        self._widget.send_js_task(
            {
                "name": "exportImage",
                "kwargs": {"resolutionScale": resolutionScale},
            }
        )

    def display_image(self):
        from IPython.display import display, Image
        import base64

        if self._widget.imageData == "":
            print(
                "No image data available, please export the image first: running export_image() in another cell."
            )
            return None
        base64_data = self._widget.imageData.split(",")[1]
        # Decode the base64 string
        image_data = base64.b64decode(base64_data)

        # Display the image
        return display(Image(data=image_data))

    def download_image(self, filename="weas-model.png"):
        self._widget.send_js_task(
            {
                "name": "tjs.downloadImage",
                "kwargs": {"filename": filename},
            }
        )

    def save_image(self, filename="weas-model.png", resolutionScale=5):
        import base64

        def _save_image():
            while not self._widget.ready:
                time.sleep(0.1)
            self.export_image(resolutionScale)
            # polling mechanism to check if the image data is available
            while not self._widget.imageData:
                time.sleep(0.1)
            base64_data = self._widget.imageData.split(",")[1]
            # Decode the base64 string
            image_data = base64.b64decode(base64_data)
            with open(filename, "wb") as f:
                f.write(image_data)

        thread = threading.Thread(target=_save_image, args=(), daemon=False)
        thread.start()



"""
# weas/src/operation/atoms.js
This module contains the Transform class which is used to perform
ReplaceOperation
AddAtomOperation
ColorByAttribute
"""


class AtomsOperation:
    def __init__(self, base_widget):
        self.base_widget = base_widget

    def replace(self, **kwargs):
        self.base_widget.send_js_task(
            {"name": "ops.atoms.ReplaceOperation", "kwargs": kwargs}
        )

    def add_atom(self, **kwargs):
        self.base_widget.send_js_task(
            {"name": "ops.atoms.AddAtomOperation", "kwargs": kwargs}
        )

    def color_by_attribute(self, **kwargs):
        self.base_widget.send_js_task(
            {"name": "ops.atoms.ColorByAttribute", "kwargs": kwargs}
        )

    def add_molecule(self, *args, **kwargs):
        from ase.build import molecule
        from weas_widget.utils import ASE_Adapter

        mol = molecule(*args, **kwargs)
        atoms = ASE_Adapter.to_weas(mol)
        self.base_widget.atoms = atoms



import numpy as np


class ASEAdapter:
    def __init__(self):
        pass

    @classmethod
    def to_weas(cls, ase_atoms):
        """Convert an ASE Atoms object to the widget's format."""
        # if atoms is a list of atoms, check if they are the same species and number of atoms
        # then convert all atoms to weas format as a list of atoms
        if isinstance(ase_atoms, list):
            if len(ase_atoms) > 0:
                for atoms in ase_atoms:
                    if (
                        atoms.get_chemical_symbols()
                        != ase_atoms[0].get_chemical_symbols()
                    ):
                        raise ValueError("All atoms must have the same species")
                weas_atoms = [cls.to_weas(atom) for atom in ase_atoms]
                return weas_atoms
            else:
                raise ValueError("The list of atoms is empty")
        species = {}
        cell = ase_atoms.get_cell().array.flatten().tolist()
        positions = ase_atoms.get_positions()
        symbols = ase_atoms.get_chemical_symbols()
        for i in range(len(symbols)):
            species[symbols[i]] = symbols[i]
        # save other arrays to attributes
        attributes = {"atom": {}, "species": {}}
        for key in ase_atoms.arrays.keys():
            if key not in ["positions", "numbers"]:
                attributes["atom"][key] = ase_atoms.arrays[key].tolist()

        weas_atoms = {
            "species": species,
            "cell": cell,
            "positions": positions.tolist(),
            "symbols": symbols,
            "attributes": attributes,
            "pbc": ase_atoms.get_pbc().tolist(),
        }
        return weas_atoms

    @classmethod
    def to_ase(cls, weas_atoms):
        # Convert the widget's format to an ASE Atoms object
        from ase import Atoms
        import numpy as np

        # if atoms is a list of atoms, convert all atoms to a list of ase atoms
        if isinstance(weas_atoms, list):
            trajectory = [cls.to_ase(atom) for atom in weas_atoms]
            return trajectory[0] if len(trajectory) == 1 else trajectory
        symbols = [weas_atoms["species"][s] for s in weas_atoms["symbols"]]
        positions = weas_atoms["positions"]
        cell = np.array(weas_atoms["cell"]).reshape(3, 3)
        ase_atoms = Atoms(
            symbols=symbols, positions=positions, cell=cell, pbc=weas_atoms["pbc"]
        )
        return ase_atoms


class PymatgenAdapter:
    def __init__(self):
        pass

    @classmethod
    def to_weas(cls, pymatgen_structure):
        from pymatgen.core import Molecule

        species = {}
        # structure is a Molecule, convert it to Structure
        if isinstance(pymatgen_structure, Molecule):
            cell = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
            pbc = [False, False, False]
        else:
            cell = pymatgen_structure.lattice.matrix.flatten().tolist()
            pbc = pymatgen_structure.lattice.pbc
        positions = [site.coords.tolist() for site in pymatgen_structure.sites]
        symbols = [site.species_string for site in pymatgen_structure.sites]
        for i in range(len(symbols)):
            species[symbols[i]] = symbols[i]
        # save other arrays to attributes
        attributes = {"atom": {}, "species": {}}
        # read pymatgen site properties
        for key in pymatgen_structure.site_properties.keys():
            attributes["atom"][key] = [
                site.properties[key] for site in pymatgen_structure.sites
            ]
        weas_atoms = {
            "species": species,
            "cell": cell,
            "positions": positions,
            "symbols": symbols,
            "attributes": attributes,
            "pbc": pbc,
        }
        return weas_atoms

    @classmethod
    def to_pymatgen(cls, weas_atoms):
        # Convert the widget's format to a Pymatgen Structure object
        from pymatgen.core import Molecule, Structure, Lattice

        if isinstance(weas_atoms, list):
            return [cls.to_pymatgen(atom) for atom in weas_atoms]
        species = weas_atoms["symbols"]
        sites = weas_atoms["positions"]
        cell = np.array(weas_atoms["cell"]).reshape(3, 3)
        # if all cell are close to zeros, it is a molecule
        if np.allclose(cell, np.zeros((3, 3))):
            structure = Molecule(species, sites)
        else:
            lattice = Lattice(weas_atoms["cell"], pbc=weas_atoms["pbc"])
            structure = Structure(lattice, species, sites, coords_are_cartesian=True)
        return structure


def load_online_example(name="tio2.cif"):
    """Load an example from the online data."""
    from ase.io import read
    import requests
    from io import StringIO

    url = "https://raw.githubusercontent.com/superstar54/weas/main/demo/datas/" + name
    # Download the file content
    response = requests.get(url)
    if response.status_code == 200:
        file_content = response.text
        # Use StringIO to simulate a file-like object for ASE to read from
        file_like_object = StringIO(file_content)
        atoms = read(file_like_object, format="cif")
        return atoms
    else:
        raise ValueError(f"Failed to download the file {name}")



import importlib.metadata
import pathlib
import anywidget
import traitlets as tl

try:
    __version__ = importlib.metadata.version("weas_widget")
except importlib.metadata.PackageNotFoundError:
    __version__ = "unknown"


class BaseWidget(anywidget.AnyWidget):
    _esm = pathlib.Path(__file__).parent / "static" / "widget.js"
    _css = pathlib.Path(__file__).parent / "static" / "widget.css"

    # indicate if the widget is displayed and available for interaction.
    ready = tl.Bool(False).tag(sync=True)
    logLevel = tl.Unicode("info").tag(sync=True)
    # atoms can be a dictionary or a list of dictionaries
    atoms = tl.Union([tl.Dict({}), tl.List(tl.Dict({}))]).tag(sync=True)
    selectedAtomsIndices = tl.List([]).tag(sync=True)
    boundary = tl.List([[0, 1], [0, 1], [0, 1]]).tag(sync=True)
    modelStyle = tl.Int(0).tag(sync=True)
    # color
    colorBy = tl.Unicode("Element").tag(sync=True)
    colorType = tl.Unicode("JMOL").tag(sync=True)
    colorRamp = tl.List(["red", "blue"]).tag(sync=True)
    # radius
    radiusType = tl.Unicode("Covalent").tag(sync=True)
    # material
    materialType = tl.Unicode("Standard").tag(sync=True)
    atomLabelType = tl.Unicode("None").tag(sync=True)
    showCell = tl.Bool(True).tag(sync=True)
    showBondedAtoms = tl.Bool(False).tag(sync=True)
    showHydrogenBonds = tl.Bool(False).tag(sync=True)
    hideLongBonds = tl.Bool(True).tag(sync=True)
    atomScales = tl.List([]).tag(sync=True)
    modelSticks = tl.List([]).tag(sync=True)
    modelPolyhedras = tl.List([]).tag(sync=True)
    volumetricData = tl.Dict({"values": [[[]]]}).tag(sync=True)
    imageData = tl.Unicode("").tag(sync=True)
    vectorField = tl.Dict().tag(sync=True)
    showVectorField = tl.Bool(True).tag(sync=True)
    guiConfig = tl.Dict({}).tag(sync=True)
    # animation
    currentFrame = tl.Int(0).tag(sync=True)
    # instanced mesh primitives
    instancedMeshPrimitive = tl.List(tl.Dict({})).tag(sync=True)
    # any mesh
    anyMesh = tl.List(tl.Dict({})).tag(sync=True)
    # viewer
    viewerStyle = tl.Dict({}).tag(sync=True)
    # camera
    cameraSetting = tl.Dict({}).tag(sync=True)
    cameraZoom = tl.Float().tag(sync=True)
    cameraPosition = tl.List().tag(sync=True)
    cameraRotation = tl.List().tag(sync=True)
    cameraLookAt = tl.List().tag(sync=True)
    # task
    js_task = tl.Dict({}).tag(sync=True)
    python_task = tl.Dict({}).tag(sync=True)
    debug = tl.Bool(False).tag(sync=True)
    # species
    speciesSettings = tl.Dict({}).tag(sync=True)
    # bond
    bondSettings = tl.Dict({}).tag(sync=True)
    # isosurface
    isoSettings = tl.Dict({}).tag(sync=True)
    # volume slice
    sliceSettings = tl.Dict({}).tag(sync=True)
    # phonon
    phonon = tl.Dict({}).tag(sync=True)
    # highlight
    highlightSettings = tl.Dict({}).tag(sync=True)
    #
    showAtomLegend = tl.Bool(False).tag(sync=True)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def send_js_task(self, task):
        """Send a task to the javascript side.
        task is a dictionary with the following keys
        - name: the name of the task
        - kwargs: a dictionary of arguments
        """
        self.js_task = {}
        self.js_task = task

    def drawModels(self):
        """Redraw the widget."""
        self.send_js_task({"name": "drawModels"})



from ..base_class import WidgetWrapper, ChangeTrackingDict
from weas_widget.data import default_bond_pairs


class BondManager(WidgetWrapper):

    catalog = "bond"

    _attribute_map = {}

    _extra_allowed_attrs = ["_settings", "settings"]

    def __init__(self, _widget):
        super().__init__(_widget)
        self._settings = ChangeTrackingDict(widget=self._widget, key="bondSettings")

    @property
    def settings(self):
        return self._settings

    @settings.setter
    def settings(self, value):
        self._settings = ChangeTrackingDict(
            value, widget=self._widget, key="bondSettings"
        )

    def update_atoms(self):
        self.settings = self.get_default_settings()

    def get_default_settings(self):
        settings = {}
        species_dict = self._widget.speciesSettings
        for specie1, data1 in species_dict.items():
            for specie2, data2 in species_dict.items():
                if (data1["element"], data2["element"]) not in default_bond_pairs:
                    continue
                bond_type = default_bond_pairs[(data1["element"], data2["element"])][2]
                min = 0
                max = (data1["radius"] + data2["radius"]) * 1.1
                color1 = data1["color"]
                color2 = data2["color"]
                if bond_type == 1:
                    min = max + 0.4
                    max = min + 1
                    color1 = "#808080"
                    color2 = "#808080"
                    bond_line_type = 1
                else:
                    bond_line_type = 0

                settings[f"{specie1}-{specie2}"] = {
                    "specie1": specie1,
                    "specie2": specie2,
                    "color1": color1,
                    "color2": color2,
                    "min": min,
                    "max": max,
                    "type": bond_line_type,
                }
        return settings



from ..base_class import WidgetWrapper, ChangeTrackingDict


class HighlightManager(WidgetWrapper):

    catalog = "highlight"

    _attribute_map = {}

    _extra_allowed_attrs = ["_settings", "settings"]

    def __init__(self, _widget):
        super().__init__(_widget)
        self._settings = ChangeTrackingDict(
            widget=self._widget, key="highlightSettings"
        )

    @property
    def settings(self):
        return self._settings

    @settings.setter
    def settings(self, value):
        self._settings = ChangeTrackingDict(
            value, widget=self._widget, key="highlightSettings"
        )

    def update_atoms(self):
        self.settings = self.get_default_settings()

    def get_default_settings(self):
        settings = {
            "selection": {
                "indices": [],
                "scale": 1.1,
                "color": "yellow",
                "type": "sphere",
            },
            "fixed": {"indices": [], "scale": 1.1, "color": "black", "type": "cross"},
        }
        return settings



from .mesh import MeshOperation
from .object import ObjectOperation
from .atoms import AtomsOperation
from .selection import SelectionOperation
from .transform import TransformOperation


class Operators:
    def __init__(self, weas_viewer):
        self.weas_viewer = weas_viewer
        self.mesh = MeshOperation(weas_viewer)
        self.object = ObjectOperation(weas_viewer)
        self.atoms = AtomsOperation(weas_viewer)
        self.selection = SelectionOperation(weas_viewer)
        self.transform = TransformOperation(weas_viewer)

    def undo(self):
        self.weas_viewer.send_js_task({"name": "ops.undo"})

    def redo(self):
        self.weas_viewer.send_js_task({"name": "ops.redo"})



from .base_class import WidgetWrapper


class Camera(WidgetWrapper):

    catalog = "camera"

    _attribute_map = {
        "zoom": "cameraZoom",
        "position": "cameraPosition",
        "look_at": "cameraLookAt",
        "setting": "cameraSetting",
    }
    _extra_allowed_attrs = []

    def __init__(self, _widget):
        super().__init__(_widget)



from .base_class import WidgetWrapper
from .plugins.vector_field import VectorField
from .plugins.isosurface import Isosurface
from .plugins.volume_slice import VolumeSlice
from .plugins.lattice_plane import LatticePlane
from .plugins.bond import BondManager
from .plugins.species import SpeciesManager
from .plugins.highlight import HighlightManager
from copy import deepcopy


class AtomsViewer(WidgetWrapper):

    catalog = "viewer"

    _attribute_map = {
        "model_style": "modelStyle",
        "selected_atoms_indices": "selectedAtomsIndices",
        "boundary": "boundary",
        "color_by": "colorBy",
        "color_ramp": "colorRamp",
        "show_cell": "showCell",
        "show_bonded_atoms": "showBondedAtoms",
        "hide_long_bonds": "hideLongBonds",
        "show_atom_legend": "showAtomLegend",
        "show_hydrogen_bonds": "showHydrogenBonds",
        "atom_label_type": "atomLabelType",
        "material_type": "materialType",
        "model_sticks": "modelSticks",
        "atom_scales": "atomScales",
        "model_polyhedras": "modelPolyhedras",
        "current_frame": "currentFrame",
        "phonon_setting": "phonon",
    }

    _extra_allowed_attrs = [
        "species",
        "vf",
        "iso",
        "volume_slice",
        "lp",
        "atoms",
        "bond",
        "highlight",
        "color_type",
    ]

    def __init__(self, _widget):

        super().__init__(_widget)
        # Initialize plugins
        setattr(self, "vf", VectorField(_widget))
        setattr(self, "iso", Isosurface(_widget))
        setattr(self, "lp", LatticePlane(_widget))
        setattr(self, "bond", BondManager(_widget))
        setattr(self, "species", SpeciesManager(_widget))
        setattr(self, "highlight", HighlightManager(_widget))
        setattr(self, "volume_slice", VolumeSlice(_widget))

    @property
    def atoms(self):
        return self._widget.atoms

    @atoms.setter
    def atoms(self, atoms):
        self._widget.atoms = atoms
        # initialize atomScales
        if isinstance(atoms, list):
            atoms = atoms[0]
        natom = len(atoms["symbols"])
        self._widget.atomScales = [1] * natom
        self._widget.modelSticks = [0] * natom
        self._widget.modelPolyhedras = [0] * natom
        # species
        self.species.update_atoms()
        # bond
        self.bond.update_atoms()
        # vector field
        self.vf.update_atoms()
        # highlight
        self.highlight.update_atoms()

    def draw(self):
        """Redraw the widget."""
        self._widget.send_js_task({"name": "avr.drawModels", "kwargs": {}})

    def set_attribute(self, name, value, domain="atoms"):
        """Set an attribute of the widget."""
        atoms = deepcopy(self._widget.atoms)
        atoms["attributes"][domain][name] = value
        self._widget.atoms = atoms
        # update the widget
        self._widget.send_js_task(
            {"name": "avr.setAttribute", "args": [name, value, domain]}
        )

    def get_attribute(self, name):
        """Get an attribute of the widget."""
        raise NotImplementedError("This method is not implemented yet.")

    @property
    def color_type(self):
        return self._widget.colorType

    @color_type.setter
    def color_type(self, value):
        self._widget.colorType = value
        self.species.update_atoms()
        self.bond.update_atoms()



from ..base_class import WidgetWrapper


class Isosurface(WidgetWrapper):

    catalog = "isosurface"

    _attribute_map = {
        "volumetric_data": "volumetricData",
        "settings": "isoSettings",
    }

    _extra_allowed_attrs = []

    def __init__(self, _widget):
        super().__init__(_widget)



"""
# weas/src/operation/transform.js
This module contains the Transform class which is used to perform
TranslateOperation
RotateOperation
ScaleOperation
"""


class TransformOperation:
    def __init__(self, base_widget):
        self.base_widget = base_widget

    def translate(self, **kwargs):
        self.base_widget.send_js_task(
            {"name": "ops.transform.TranslateOperation", "kwargs": kwargs}
        )

    def rotate(self, **kwargs):
        self.base_widget.send_js_task(
            {"name": "ops.transform.RotateOperation", "kwargs": kwargs}
        )

    def scale(self, **kwargs):
        self.base_widget.send_js_task(
            {"name": "ops.transform.ScaleOperation", "kwargs": kwargs}
        )
